///
///    This file is part of ILNumerics Community Edition.
///
///    ILNumerics Community Edition - high performance computing for applications.
///    Copyright (C) 2006 - 2012 Haymo Kutschbach, http://ilnumerics.net
///
///    ILNumerics Community Edition is free software: you can redistribute it and/or modify
///    it under the terms of the GNU General Public License version 3 as published by
///    the Free Software Foundation.
///
///    ILNumerics Community Edition is distributed in the hope that it will be useful,
///    but WITHOUT ANY WARRANTY; without even the implied warranty of
///    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
///    GNU General Public License for more details.
///
///    You should have received a copy of the GNU General Public License
///    along with ILNumerics Community Edition. See the file License.txt in the root
///    of your distribution package. If not, see <http://www.gnu.org/licenses/>.
///
///    In addition this software uses the following components and/or licenses: 
///
///    =================================================================================
///    The Open Toolkit Library License
///    
///    Copyright (c) 2006 - 2009 the Open Toolkit library.
///    
///    Permission is hereby granted, free of charge, to any person obtaining a copy
///    of this software and associated documentation files (the "Software"), to deal
///    in the Software without restriction, including without limitation the rights to 
///    use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
///    the Software, and to permit persons to whom the Software is furnished to do
///    so, subject to the following conditions:
///
///    The above copyright notice and this permission notice shall be included in all
///    copies or substantial portions of the Software.
///
///    =================================================================================
///    

using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices; 
using ILNumerics.Storage;
using ILNumerics.Misc;
using ILNumerics.Native;
using ILNumerics.Exceptions;

namespace ILNumerics {

	public partial class ILMath {

        [ThreadStatic]
        private static Random m_randomGenerator;
        /// <summary>
        /// Random number generator used for any generation of random numbers.
        /// </summary>
        /// <remarks>This property is thread safe (by using the [ThreadStatic] attribute).</remarks>
        protected static Random RandomGenerator {
            get {
                if (m_randomGenerator == null) {
                    m_randomGenerator = new Random(Environment.TickCount + System.Threading.Thread.CurrentThread.ManagedThreadId);
                }
                return m_randomGenerator;
            }
        }
        /// <summary>
        /// Pseudo random N-dimensional array elements.
        /// </summary>
        /// <param name="dimensions">int array or single int paramters specifying 
        /// dimensions for new array to be created.</param>
        /// <returns>N-dimensional array filled with random numbers.</returns>
        /// <remarks><para>The elements lie within the range 0.0 ... 1.0 and are uniformly 
        /// distributed.</para>
        /// <para>The initial seed will be set to Environment.TickCount on the first call.</para></remarks>
        public static ILRetArray<double> rand(params int[] dimensions) {
            ILSize dim = new ILSize(dimensions);
            double[] data = ILMemoryPool.Pool.New<double>(dim.NumberOfElements);
            if (dim.NumberOfElements < ILNumerics.Settings.s_minParallelElement1Count) {
                unsafe {
                    fixed (double* pRetArray = data) {
                        double* pCurIdx = pRetArray;
                        double* pLastElement = pCurIdx + dim.NumberOfElements;
                        while (pCurIdx < pLastElement)
                            *pCurIdx++ = RandomGenerator.NextDouble();
                    }
                }
            } else {
                System.Threading.Tasks.Parallel.For(0, dim.NumberOfElements, i => {
                    data[i] = RandomGenerator.NextDouble();
                });
            }
            return new ILRetArray<double>(data, dim);
        }

    }
}
